# -*- coding: utf-8 -*-

from __future__ import annotations

import copy
import json
from collections.abc import Mapping
from typing import Optional, Type, TypeVar

import pytoml
from ruyaml import YAML


class Params(dict):
    T = TypeVar("T", bound="Params")

    def copy(self) -> T:
        return copy.deepcopy(self)

    def merge(self, default: Mapping) -> None:
        for key, value in copy.deepcopy(default).items():
            self.setdefault(key, value)

    def to_json_string(self, **kwargs) -> str:
        if not kwargs:
            kwargs = {"ensure_ascii": False}

        return json.dumps(self, **kwargs)

    def to_json(
        self, filename: str, file_kwargs: Optional[Mapping] = None, **kwargs
    ) -> None:
        if not file_kwargs:
            file_kwargs = {"mode": "w"}
        kwargs.setdefault("ensure_ascii", False)
        kwargs.setdefault("indent", 4)

        with open(filename, **file_kwargs) as f:
            json.dump(self, f, **kwargs)

    dumps = to_json_string
    dump = to_json

    def to_yaml(
        self, filename: str, file_kwargs: Optional[Mapping] = None, **kwargs
    ) -> None:
        if not file_kwargs:
            file_kwargs = {"mode": "w"}

        if not kwargs:
            kwargs = {}

        with open(filename, **file_kwargs) as f:
            YAML().dump(dict(self), f, **kwargs)

    def to_toml(
        self, filename: str, file_kwargs: Optional[Mapping] = None, **kwargs
    ) -> None:
        if not file_kwargs:
            file_kwargs = {"mode": "w"}

        with open(filename, **file_kwargs) as f:
            pytoml.dump(self, f, **kwargs)

    @classmethod
    def from_json_string(cls: Type[T], json_string: str, **kwargs) -> T:
        return cls(json.loads(json_string, **kwargs))

    @classmethod
    def from_json(
        cls: Type[T], filename: str, file_kwargs: Optional[Mapping] = None, **kwargs
    ) -> T:
        if not file_kwargs:
            file_kwargs = {"mode": "r"}

        with open(filename, **file_kwargs) as f:
            return cls(json.load(f, **kwargs))

    loads = from_json_string
    load = from_json

    @classmethod
    def from_toml(
        cls: Type[T], filename: str, file_kwargs: Optional[Mapping] = None, **kwargs
    ) -> T:
        if not file_kwargs:
            file_kwargs = {"mode": "r"}

        with open(filename, **file_kwargs) as f:
            return cls(pytoml.load(f, **kwargs))

    @classmethod
    def from_yaml(
        cls: Type[T], filename: str, file_kwargs: Optional[Mapping] = None, **kwargs
    ) -> T:
        if not file_kwargs:
            file_kwargs = {"mode": "r"}

        with open(filename, **file_kwargs) as f:
            return cls(YAML(typ="safe").load(f, **kwargs))



# -*- coding: utf-8 -*-

from __future__ import annotations

import statistics
from collections.abc import Iterable, Sequence
from typing import Any, Optional, Union

import numpy as np


def describe_series(
    s: Sequence, r: int = 2, qs: Optional[Sequence[float]] = None
) -> dict:
    if not qs:
        qs = [10, 25, 50, 75, 90, 95, 99, 99.9, 99.99]

    info = {
        "size": len(s),
        "mode": statistics.mode(s),
        "mean": float(np.mean(s).round(r)),
        "std": float(np.std(s).round(r)),
        "min": float(np.min(s)),
        "max": float(np.max(s)),
    }

    percentiles = np.percentile(s, qs).round(r)
    for q, p in zip(qs, percentiles):
        info.update({f"{q}%": p.round(r)})

    return info


def split(
    *data: Iterable[Sequence],
    val_size: float = 0.1,
    test_size: float = 0.2,
    random_state: Optional[int] = None,
    shuffle: bool = True,
    stratify: Optional[Sequence] = None,
) -> Union[dict[str, Any], dict[str, tuple]]:
    # pylint: disable=import-outside-toplevel
    from sklearn.model_selection import train_test_split

    def _split(data, size, stratify=None):
        params = data + (stratify,) if stratify is not None else data
        splits = train_test_split(
            *params,
            test_size=size,
            random_state=random_state,
            shuffle=shuffle,
            stratify=stratify,
        )

        split0, split1 = tuple(splits[::2]), tuple(splits[1::2])

        return split0, split1

    def _format_output(x):
        if stratify:
            x = x[:-1]

        return x if len(x) > 1 else x[0]

    val_size /= 1 - test_size
    train_val, test = _split(data, test_size, stratify=stratify)

    if stratify is None:
        train, val = _split(train_val, val_size, stratify=None)
    else:
        *train_val, train_val_stratify = train_val
        train, val = _split(tuple(train_val), val_size, stratify=train_val_stratify)

    return {
        "train": _format_output(train),
        "val": _format_output(val),
        "test": _format_output(test),
    }



# -*- coding: utf-8 -*-

from __future__ import annotations

import contextlib
import importlib
import random
from collections.abc import Iterator
from typing import Optional, Union

from carton.palette import Colors


def random_state(
    seed: Optional[int] = None, low: int = 0, high: int = 2 ** 32 - 1
) -> int:
    if seed is not None:
        return seed

    return random.randint(low, high)


def _import_module(module):
    try:
        module = importlib.import_module(module)
    except ModuleNotFoundError:
        return None

    return module


def set_seed(seed: int, debug: bool = False) -> tuple:
    states: tuple = (random.getstate(),)

    random.seed(seed)

    numpy = _import_module("numpy")
    if numpy:
        states += (numpy.random.get_state(),)
        numpy.random.seed(seed)

    torch = _import_module("torch")
    if torch:
        cuda = torch.cuda.is_available()

        states += (
            torch.get_rng_state(),
            torch.cuda.get_rng_state() if cuda else None,
            torch.cuda.get_rng_state_all() if cuda else None,
        )
        torch.manual_seed(seed)
        torch.random.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

        if debug:
            states += (
                torch.backends.cudnn.enabled,
                torch.backends.cudnn.benchmark,
                torch.backends.cudnn.deterministic,
            )
            torch.backends.cudnn.enabled = False
            torch.backends.cudnn.benchmark = False
            torch.backends.cudnn.deterministic = True

    return states


@contextlib.contextmanager
def seed(random_seed: int) -> Iterator[None]:
    state = random.getstate()
    random.seed(random_seed)
    try:
        yield
    finally:
        random.setstate(state)


def random_string(
    length: int = 10,
    chars: str = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
) -> str:
    return "".join(random.choices(chars, k=length))


def random_colors(
    colors: Optional[list[str]] = None,
    num: int = 1,
    replacement: bool = False,
) -> Union[str, list[str]]:
    if colors is None:
        colors = Colors.PRESETS

    sample_function = random.choices if replacement else random.sample
    selected = sample_function(colors, k=num)

    return selected if num > 1 else selected[0]



# -*- coding: utf-8 -*-

from __future__ import annotations

import logging
import os
import sys
from collections.abc import Mapping
from typing import Optional, Union


def _setup_handlers(logger, handlers=None, reset=True):
    if reset:
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)
            handler.close()

    if handlers:
        for handler in handlers:
            logger.addHandler(handler)


def _setup_formaters(logger, format_):
    for handler in logger.handlers:
        handler.setFormatter(logging.Formatter(format_))


def _setup_logger(logger, level, format_, handlers):
    logger.setLevel(level)
    _setup_handlers(logger, handlers, reset=True)
    _setup_formaters(logger, format_)


#  pylint: disable=redefined-builtin
def setup_logger(
    logger: Optional[logging.Logger] = None,
    level: Optional[Union[int, str]] = logging.INFO,
    format: str = "%(asctime)s %(name)s %(levelname)s: %(message)s",
    force: bool = True,
    stream: bool = True,
    filename: Optional[Union[str, bytes, os.PathLike]] = None,
    **kwargs,
) -> None:
    handlers = []
    if stream:
        handlers += [logging.StreamHandler()]

    if filename:
        handlers += [logging.FileHandler(filename)]

    if logger is not None:
        _setup_logger(logger, level, format, handlers)
        return

    kwargs = {"level": level, "format": format, "handlers": handlers}
    if force:
        if sys.version_info < (3, 8):
            _setup_handlers(logging.root, reset=True)
        else:
            kwargs["force"] = force

    logging.basicConfig(**kwargs)


def log_dict(_logger: logging.Logger, d: Mapping, sep: str = " = ") -> None:
    for key, value in d.items():
        _logger.info(f"{key}{sep}{value}")



# -*- coding: utf-8 -*-

from __future__ import annotations

import contextlib
import os
import sys
from collections.abc import Iterator


@contextlib.contextmanager
def suppress_stdout(suppress: bool = True) -> Iterator[None]:
    with open(os.devnull, mode="w") as devnull:
        stdout = sys.stdout
        if suppress:
            sys.stdout = devnull
        try:
            yield
        finally:
            if suppress:
                sys.stdout = stdout



# -*- coding: utf-8 -*-

from __future__ import annotations

from collections.abc import Iterable
from typing import Union

try:
    import pandas as pd
    import plotly.graph_objects as go
    import streamlit as st
    from pandas import DataFrame
except ModuleNotFoundError:
    raise RuntimeError(
        "Please install `pandas`, `streamlit` and `plotly` library to use gallary."
    )


def display_clustering(
    prediction_files: Iterable[str],
    df_score: Union[str, DataFrame],
    limit: int = 100,
    k_field: str = "k",
    score_fields: Union[str, Iterable[str]] = "score",
    width: int = 1000,
) -> None:
    if isinstance(df_score, str):
        df_score = pd.read_csv(df_score)

    if isinstance(score_fields, str):
        score_fields = [score_fields]
    else:
        score_fields = list(score_fields)  # type: ignore

    # Sidebar
    st.set_page_config(layout="wide")

    st.sidebar.header("Files")
    filename = st.sidebar.selectbox("Select a file", list(prediction_files))
    df = pd.read_csv(filename)

    st.sidebar.header("Label")
    label = st.sidebar.selectbox("Select a label", sorted(df["prediction"].unique()))

    filtered = df[df["prediction"] == label]

    st.sidebar.header("Options")
    limit = st.sidebar.slider(
        "Numers of examples", min_value=1, max_value=len(filtered), value=limit
    )
    hide_overview = st.sidebar.checkbox("Hide overview", False)

    # Body
    st.title("Clustering Report")
    st.header("Overview")

    # Overview
    if not hide_overview:
        w = 0.1
        fig = go.Figure()
        for i, field in enumerate(score_fields):
            fig.add_trace(
                go.Scatter(
                    x=df_score[k_field],
                    y=df_score[field],
                    name=field,
                    yaxis=f"y{i + 1}" if i > 0 else "y",
                )
            )

            layout = {
                "title": field,
            }
            if i > 0:
                layout.update(
                    {
                        "title": field,
                        "overlaying": "y",
                        "side": "right" if i % 2 == 1 else "left",
                        "position": 1 - (i // 2) * w if i % 2 == 1 else (i // 2) * w,
                        "anchor": "x" if i == 1 else "free",
                    }
                )

            fig.update_layout(
                {
                    f'yaxis{i + 1 if i > 0 else ""}': layout,
                }
            )
        fig.update_layout(
            {
                "xaxis": {
                    "domain": [
                        ((len(score_fields) + 1) // 2) * w,
                        1 - (len(score_fields) // 2) * w,
                    ],
                },
                "width": width,
            }
        )
        st.plotly_chart(fig)

    # Predictions
    st.header("Predictions")
    st.table(filtered.iloc[:limit])



# -*- coding: utf-8 -*-

from __future__ import annotations

import glob
import itertools
import math
import multiprocessing
import os
from collections.abc import Callable, Generator, Iterable
from typing import Any, Optional, Union


def normalize_path(p: str) -> str:
    return os.path.abspath(os.path.expanduser(p))


class _Path(str):
    def __new__(cls, value):
        return super().__new__(cls, (value))

    def __truediv__(self, path):
        return _Path(os.path.join(self, path))

    def norm(self):
        return self.__class__(normalize_path(self))


def path(root: str) -> _Path:
    return _Path(root)


def iter_file_groups(
    dirname: str,
    exts: Union[str, Iterable[str]],
    with_key: bool = False,
    missing: str = "error",
) -> Union[
    Iterable[str], Iterable[tuple[str, ...]], tuple[str, Iterable[tuple[str, ...]]]
]:
    def _format_ext(ext):
        return f'.{ext.lstrip(".")}'

    def _iter_files(dirname):
        for dirpath, _, filenames in os.walk(dirname):
            for filename in filenames:
                if os.path.splitext(filename)[1] in exts:
                    yield os.path.join(dirpath, filename)

    missings = {"error", "ignore"}
    if missing not in missings:
        raise ValueError(f"Param `missing` should be in {missings}")

    if isinstance(exts, str):
        return glob.iglob(
            os.path.join(dirname, f"**/*{_format_ext(exts)}"), recursive=True
        )

    exts = {*map(_format_ext, exts)}
    num_exts = len(exts)
    files = _iter_files(dirname)
    for key, group in itertools.groupby(
        sorted(files), key=lambda x: os.path.splitext(os.path.relpath(x, dirname))[0]
    ):
        sorted_group = sorted(group, key=lambda x: os.path.splitext(x)[1])
        if len(sorted_group) != num_exts and missing == "error":
            raise RuntimeError(f"Missing files: {key}.{exts}")

        yield (key, sorted_group) if with_key else sorted_group


def read_lines(filename: str, chunk_size: int = 64 * 1024) -> Iterable[str]:
    with open(filename, mode="r") as f:
        remaining = ""
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break

            linefeed = chunk.rfind("\n")
            if linefeed > 0:
                chunk, remaining = remaining + chunk[:linefeed], chunk[linefeed + 1 :]
            else:
                remaining = ""

            for line in chunk.split("\n"):
                yield line


def iter_lines(
    filename: str,
    ignore_empty: bool = True,
    strip: bool = True,
    transform: Callable[[str], str] = lambda x: x,
) -> Iterable[str]:
    with open(filename, mode="r") as f:
        for line in f:
            if strip:
                line = line.rstrip()

            if line or not ignore_empty:
                yield transform(line)


def _get_chunk_size(filename, num_workers=multiprocessing.cpu_count()):
    return math.ceil(os.path.getsize(filename) / num_workers)


def _get_chunkified_args(
    filename, *args, num_workers=multiprocessing.cpu_count(), chunk_size=None
):
    if not chunk_size:
        chunk_size = _get_chunk_size(filename, num_workers=num_workers)

    return [
        (filename, *args, start, end)
        for filename, *args, (start, end) in zip(
            itertools.repeat(filename),
            *[itertools.repeat(arg) for arg in args],
            chunkify(filename, chunk_size=chunk_size),
        )
    ]


def _iter_line_wrapper(filename, fn, start, end):
    return [fn(line) for line in readlines(filename, start, end)]


def chunkify(
    filename: Union[str, os.PathLike], chunk_size: int = 1024 * 1024
) -> Generator[tuple[int, int], None, None]:
    start = 0
    size = os.path.getsize(filename)
    with open(filename, mode="rb") as f:
        while True:
            f.seek(chunk_size, 1)
            f.readline()
            end = min(f.tell(), size)
            yield start, end

            if end >= size:
                break
            start = end


def readlines(
    filename: Union[str, os.PathLike],
    start: Optional[int] = None,
    end: Optional[int] = None,
    **kwargs,
) -> Generator[str, None, None]:
    with open(filename, **kwargs) as f:
        if start:
            f.seek(start)
        while True:
            line = f.readline()
            if end and f.tell() > end or not line:
                break

            yield line


def map_text(
    filename: Union[str, os.PathLike],
    fn: Callable[[str, int, int], Any],
    num_workers: int = multiprocessing.cpu_count(),
    chunk_size: Optional[int] = None,
) -> Iterable:
    args = _get_chunkified_args(
        filename, num_workers=num_workers, chunk_size=chunk_size
    )
    with multiprocessing.Pool(processes=num_workers) as p:
        data = p.starmap(fn, args)

    return itertools.chain.from_iterable(data)


def map_lines(
    filename: Union[str, os.PathLike],
    fn: Callable[[str], Any],
    num_workers: int = multiprocessing.cpu_count(),
    chunk_size: Optional[int] = None,
) -> Iterable:
    args = _get_chunkified_args(
        filename, fn, num_workers=num_workers, chunk_size=chunk_size
    )
    with multiprocessing.Pool(processes=num_workers) as p:
        data = p.starmap(_iter_line_wrapper, args)

    return itertools.chain.from_iterable(data)



# -*- coding: utf-8 -*-

from __future__ import annotations

import collections
import itertools
import warnings
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import Any, Hashable, Optional, Union

from carton.utils import identity


def iterable(x: Any) -> bool:
    return isinstance(x, collections.abc.Iterable) and not isinstance(x, (str, bytes))


def dict_to_tuple(d: Mapping, keys: Optional[Iterable[Hashable]] = None) -> tuple:
    if not keys:
        return tuple(d.values())

    return tuple(d[x] for x in keys)


def get_dict_to_tuple_function(keys: Iterable[Hashable]) -> Callable[[Mapping], tuple]:
    keys = list(keys)

    def f(d):
        return dict_to_tuple(d, keys)

    return f


def flatten_dict(d: Mapping, sep: str = ".") -> dict[str, Any]:
    flattened = {}

    def _flatten(d, prefix):
        for key, value in d.items():
            key = str(key)
            if sep in key:
                raise ValueError(f"`sep` in `key` ({key!r}) is not allowed")

            new_prefix = sep.join([prefix, key]) if prefix else key
            if not isinstance(value, collections.abc.Mapping):
                flattened[new_prefix] = value
            else:
                _flatten(value, new_prefix)

    _flatten(d, "")

    return flattened


def roughen_dict(d: Mapping[str, Any], sep: str = ".") -> dict[str, Any]:
    roughen = {}
    for key, value in d.items():
        if isinstance(value, collections.abc.Mapping):
            warnings.warn(f"Nested mapping found: {key!r}")

        if sep not in key:
            roughen[key] = value
            continue

        *keys, last_key = key.split(sep)
        child = roughen
        for subkey in keys:
            child = child.setdefault(subkey, {})
        child[last_key] = value

    return roughen


def chain_get(d: Mapping, keys: Union[str, Sequence[str]], sep=".") -> Any:
    if isinstance(keys, str):
        if sep not in keys:
            return d[keys]
        keys = keys.split(sep)

    for key in keys:
        d = d[key]

    return d


def collate(
    data: Iterable[Mapping],
    keys: Optional[Sequence[Hashable]] = None,
    collate_fn: Optional[Union[Callable, Mapping[Hashable, Callable]]] = None,
) -> dict:
    def _get_collate_fn(key=None):
        if isinstance(collate_fn, collections.abc.Mapping):
            return collate_fn.get(key, identity)

        return collate_fn

    it = iter(data)

    first: Mapping = next(it, {})
    first_keys = first.keys()

    if not keys:
        keys = list(first_keys)
    elif not iterable(keys):
        keys = [keys]

    collated: dict = {}
    for x in itertools.chain((first,), it):
        if not isinstance(x, collections.abc.Mapping):
            raise TypeError(f"Not a mapping: {x}")

        if first_keys != x.keys():
            raise ValueError(f"Inconsistent keys:{first_keys!r} != {x.keys()!r}")

        for key in keys:
            collated.setdefault(key, []).append(x[key])

    if collate_fn is not None:
        collated = {key: _get_collate_fn(key)(value) for key, value in collated.items()}

    if len(collated) > 1 or len(collated) == 0:
        return collated

    return next(iter(collated.values()))


def chunk(it: Iterable, size: int) -> Iterable:
    # Credit: https://stackoverflow.com/a/22045226/1831512
    it = iter(it)

    return iter(lambda: tuple(itertools.islice(it, size)), ())



# -*- coding: utf-8 -*-

__all__ = [
    "collections.py",
    "data.py",
    "datetime.py",
    "file.py",
    "gallary.py",
    "io.py",
    "logger.py",
    "palette.py",
    "params.py",
    "random.py",
    "utils.py",
]


__version__ = "0.3.6"
